
/* 
 * Codifica di Huffman (6/06/2021)
 *
 * Classe Huffman: algoritmi di compressione e decompressione
 */

// import java.util.*;

import campus_toolkit.*;


public class HuffmanDecompression {


  private static final int BYTES = InputByteFile.BYTES;
  
  
  // ---------  Decompressione  ---------
  
  
  // 6.
  // Ricostruzione dell'albero di Huffman dalla sua codifica lineare
  // in: documento compresso
  // Struttura:
  // - una foglia è rappresentata dal carattere corrispondente
  // - un albero con più di un nodo è rappresentato
  //   dalla linearizzazione dei sottoalberi sinistro e destro,
  //   precedute dal simbolo '@'
  
  public static Node restoreTree( InputByteFile in ) {
  
    int b = in.readByte();                                  // byte dell'intestazione
    
    if ( b == '@' ) {                                       // '@' ?
    
      Node left  = restoreTree( in );                       // sottoalbero sinistro
      
      Node right = restoreTree( in );                       // sottoalbero destro
      
      return new Node( left, right );                       // nodo genitore
    
    } else {
      if ( b == '\\' ) {
        b = in.readByte();                                  // carattere speciale (skip)
      }
      return new Node( b, 0 );                              // foglia
    }
  }
  
  
  // 7.
  // Decompressione
  // src: nome del documento compresso
  // dst: nome del documento ripristinato
  
  public static void decompress( String src, String dst ) {
    
    InputByteFile in = new InputByteFile( src );            // documento compresso
    
    int count = restoreCount( in );                         // decodifica dell'intestazione
    int dummy = in.readByte();                              // 'fine linea'
    Node root = restoreTree( in );
    dummy = in.readByte();                                  // codici caratteri dopo il 'fine linea'
    
    OutputByteFile out = new OutputByteFile( dst );         // documento da ripristinare
    
    for ( int j=0; j<count; j=j+1 ) {                       // scansione: decodifica dei caratteri
      
      int b = decodeNextByte( root, in );
      out.writeByte( b );
    }
    in.close();
    out.close();
  }
  
  private static int restoreCount( InputByteFile in ) {
  
    int count = 0;
    for ( int i=0; i<4; i=i+1 ) {                           // lunghezza del documento < maxint
    
      count = BYTES * count + in.readByte();
    }
    return count;
  }
  
  // Decodifica del carattere successivo
  // root: nodo radice dell'albero di Huffman
  // in:   documento compresso
  
  private static int decodeNextByte( Node root, InputByteFile in ) {
  
    Node n = root;                                          // percorso lungo l'albero di Huffman
    
    do {
      if ( in.readBit() == 0 ) {
        n = n.left();                                       // bit 0: figlio sinistro
      } else {
        n = n.right();                                      // bit 1: figlio destro
      }
    } while ( !n.isLeaf() );                                // fino al raggiungimento di una foglia
    
    return n.getByte();                                     // byte individuato
  }
  
  
}  // class HuffmanDecompression

